#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""Diode SDK Client for Orb Discovery."""

import logging
import threading

from netboxlabs.diode.sdk import DiodeClient

from device_discovery.translate import translate_data
from device_discovery.version import version_semver

APP_NAME = "device-discovery"
APP_VERSION = version_semver()

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Client:
    """
    Singleton class for managing the Diode client for device-discovery.

    This class ensures only one instance of the Diode client is created and provides methods
    to initialize the client and ingest data.

    Attributes
    ----------
        diode_client (DiodeClient): Instance of the DiodeClient.

    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        """
        Create a new instance of the Client if one does not already exist.

        Returns
        -------
            Client: The singleton instance of the Client.

        """
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        """Initialize the Client instance with no Diode client."""
        if not hasattr(self, "diode_client"):  # Prevent reinitialization
            self.diode_client = None

    def init_client(self, prefix: str, target: str, api_key: str | None = None):
        """
        Initialize the Diode client with the specified target, API key, and TLS verification.

        Args:
        ----
            prefix (str): The prefix for the producer app name.
            target (str): The target endpoint for the Diode client.
            api_key (Optional[str]): The API key for authentication (default is None).

        """
        with self._lock:
            self.diode_client = DiodeClient(
                target=target,
                app_name=f"{prefix}/{APP_NAME}" if prefix else APP_NAME,
                app_version=APP_VERSION,
                api_key=api_key,
            )

    def ingest(self, hostname: str, data: dict):
        """
        Ingest data using the Diode client after translating it.

        Args:
        ----
            hostname (str): The device hostname.
            data (dict): The data to be ingested.

        Raises:
        ------
            ValueError: If the Diode client is not initialized.

        """
        if self.diode_client is None:
            raise ValueError("Diode client not initialized")

        with self._lock:
            response = self.diode_client.ingest(translate_data(data))

        if response.errors:
            logger.error(f"ERROR ingestion failed for {hostname} : {response.errors}")
        else:
            logger.info(f"Hostname {hostname}: Successful ingestion")



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""NetBox Labs - Translate Unit Tests."""

import pytest
from netboxlabs.diode.sdk.ingester import Tag

from device_discovery.policy.models import Defaults, ObjectParameters
from device_discovery.translate import (
    translate_data,
    translate_device,
    translate_interface,
    translate_interface_ips,
)


@pytest.fixture
def sample_device_info():
    """Sample device information for testing."""
    return {
        "hostname": "router1",
        "model": "ISR4451",
        "vendor": "Cisco",
        "serial_number": "123456789",
        "site": "New York",
        "driver": "ios",
        "interface_list": ["GigabitEthernet0/0", "GigabitEthernet0/0/1"],
    }


@pytest.fixture
def sample_interface_info():
    """Sample interface information for testing."""
    return {
        "GigabitEthernet0/0": {
            "is_enabled": True,
            "mtu": 1500,
            "mac_address": "00:1C:58:29:4A:71",
            "speed": 1000,
            "description": "Uplink Interface",
        },
        "GigabitEthernet0/0/1": {
            "is_enabled": True,
            "mtu": 1500,
            "mac_address": "00:1C:58:29:4A:72",
            "speed": 10000,
            "description": "Uplink Interface",
        }
    }


@pytest.fixture
def sample_interface_overflows_info():
    """Sample interface information for testing."""
    return {
        "GigabitEthernet0/0": {
            "is_enabled": True,
            "mtu": 150000000000,
            "mac_address": "00:1C:58:29:4A:71",
            "speed": 10000000000,
            "description": "Uplink Interface",
        }
    }


@pytest.fixture
def sample_interfaces_ip():
    """Sample interface IPs for testing."""
    return {"GigabitEthernet0/0/1": {"ipv4": {"192.0.2.1": {"prefix_length": 24}}}}


@pytest.fixture
def sample_defaults():
    """Sample defaults for testing."""
    return Defaults(
        site="New York",
        tags=["tag1", "tag2"],
        device=ObjectParameters(comments="testing", tags=["devtag"]),
        interface=ObjectParameters(description="testing", tags=["inttag"]),
        ipaddress=ObjectParameters(description="ip test", tags=["iptag"]),
        prefix=ObjectParameters(description="prefix test",tags=["prefixtag"]),
        role="router",
    )


def test_translate_device(sample_device_info, sample_defaults):
    """Ensure device translation is correct."""
    device = translate_device(sample_device_info, sample_defaults)
    assert device.name == "router1"
    assert device.device_type.model == "ISR4451"
    assert device.platform.name == "ios"
    assert device.serial == "123456789"
    assert device.site.name == "New York"
    assert device.comments == "testing"
    assert device.tags == [Tag(name="tag1"), Tag(name="tag2"), Tag(name="devtag")]


def test_translate_interface(
    sample_device_info, sample_interface_info, sample_defaults
):
    """Ensure interface translation is correct."""
    device = translate_device(sample_device_info, sample_defaults)
    interface = translate_interface(
        device,
        "GigabitEthernet0/0",
        sample_interface_info["GigabitEthernet0/0"],
        sample_defaults,
    )
    assert interface.device.name == "router1"
    assert interface.name == "GigabitEthernet0/0"
    assert interface.enabled is True
    assert interface.mtu == 1500
    assert interface.mac_address == "00:1C:58:29:4A:71"
    assert interface.speed == 1000000
    assert interface.description == "Uplink Interface"
    assert interface.tags == [Tag(name="tag1"), Tag(name="tag2"), Tag(name="inttag")]


def test_translate_interface_with_overflow_data(
    sample_device_info, sample_interface_overflows_info, sample_defaults
):
    """Ensure interface translation is correct."""
    device = translate_device(sample_device_info, sample_defaults)
    interface = translate_interface(
        device,
        "GigabitEthernet0/0",
        sample_interface_overflows_info["GigabitEthernet0/0"],
        sample_defaults,
    )
    assert interface.device.name == "router1"
    assert interface.name == "GigabitEthernet0/0"
    assert interface.enabled is True
    assert interface.mtu == 0
    assert interface.mac_address == "00:1C:58:29:4A:71"
    assert interface.speed == 0
    assert interface.description == "Uplink Interface"
    assert interface.tags == [Tag(name="tag1"), Tag(name="tag2"), Tag(name="inttag")]


def test_translate_interface_ips(
    sample_device_info, sample_interface_info, sample_interfaces_ip, sample_defaults
):
    """Ensure interface IPs translation is correct."""
    device = translate_device(sample_device_info, sample_defaults)
    interface = translate_interface(
        device,
        "GigabitEthernet0/0",
        sample_interface_info["GigabitEthernet0/0"],
        sample_defaults,
    )
    ip_entities = list(
        translate_interface_ips(interface, sample_interfaces_ip, sample_defaults)
    )

    assert len(ip_entities) == 0

    interface = translate_interface(
        device,
        "GigabitEthernet0/0/1",
        sample_interface_info["GigabitEthernet0/0/1"],
        sample_defaults,
    )
    ip_entities = list(
        translate_interface_ips(interface, sample_interfaces_ip, sample_defaults)
    )

    assert len(ip_entities) == 2
    assert ip_entities[0].prefix.prefix == "192.0.2.0/24"
    assert ip_entities[1].ip_address.address == "192.0.2.1/24"
    assert ip_entities[0].prefix.description == "prefix test"
    assert ip_entities[1].ip_address.description == "ip test"
    assert ip_entities[0].prefix.tags == [Tag(name="tag1"), Tag(name="tag2"), Tag(name="prefixtag")]
    assert ip_entities[1].ip_address.tags == [Tag(name="tag1"), Tag(name="tag2"), Tag(name="iptag")]


def test_translate_data(
    sample_device_info, sample_interface_info, sample_interfaces_ip
):
    """Ensure data translation is correct."""
    data = {
        "device": sample_device_info,
        "interface": sample_interface_info,
        "interface_ip": sample_interfaces_ip,
        "driver": "ios",
    }
    entities = list(translate_data(data))
    assert len(entities) == 5
    assert entities[0].device.name == "router1"
    assert entities[1].interface.name == "GigabitEthernet0/0"
    assert entities[2].interface.name == "GigabitEthernet0/0/1"
    assert entities[3].prefix.prefix == "192.0.2.0/24"
    assert entities[4].ip_address.address == "192.0.2.1/24"



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""Device Discovery Policy Models."""

from enum import Enum
from typing import Any

from croniter import CroniterBadCronError, croniter
from pydantic import BaseModel, Field, field_validator


class Status(Enum):
    """Enumeration for status."""

    NEW = "new"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"


class Napalm(BaseModel):
    """Model for NAPALM configuration."""

    driver: str | None = Field(default=None, description="Driver name, optional")
    hostname: str
    username: str
    password: str
    timeout: int = 60
    optional_args: dict[str, Any] | None = Field(
        default=None, description="Optional arguments"
    )

class ObjectParameters(BaseModel):
    """Model for object parameters."""

    comments: str | None = Field(default=None, description="Comments, optional")
    description: str | None = Field(default=None, description="Description, optional")
    tags: list[str] | None = Field(default=None, description="Tags, optional")

class Defaults(BaseModel):
    """Model for default configuration."""

    site: str | None = Field(default=None, description="Site name, optional")
    role: str | None = Field(default=None, description="Device Role name, optional")
    tags: list[str] | None = Field(default=None, description="Tags, optional")
    device: ObjectParameters | None = Field(default=None, description="Device parameters, optional")
    interface: ObjectParameters | None = Field(default=None, description="Interface parameters, optional")
    ipaddress: ObjectParameters | None = Field(default=None, description="IP Address parameters, optional")
    prefix: ObjectParameters | None = Field(default=None, description="Prefix parameters, optional")

class Config(BaseModel):
    """Model for discovery configuration."""

    schedule: str | None = Field(default=None, description="cron interval, optional")
    defaults: Defaults | None = Field(default=None, description="Default configuration, optional")

    @field_validator("schedule")
    @classmethod
    def validate_cron(cls, value):
        """
        Validate the cron schedule format.

        Args:
        ----
            value: The cron schedule value.

        Raises:
        ------
            ValueError: If the cron schedule format is invalid.

        """
        try:
            croniter(value)
        except CroniterBadCronError:
            raise ValueError("Invalid cron schedule format.")
        return value


class Policy(BaseModel):
    """Model for a policy configuration."""

    config: Config | None = Field(default=None, description="Configuration data")
    scope: list[Napalm]


class PolicyRequest(BaseModel):
    """Model for a policy request."""

    policies: dict[str, Policy]



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""Device Discovery Policy Runner."""

import logging
import uuid
from datetime import datetime, timedelta

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from napalm import get_network_driver

from device_discovery.client import Client
from device_discovery.discovery import discover_device_driver, supported_drivers
from device_discovery.policy.models import Config, Defaults, Napalm, Status

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PolicyRunner:
    """Policy Runner class."""

    def __init__(self):
        """Initialize the PolicyRunner."""
        self.name = ""
        self.scopes = dict[str, Napalm]()
        self.config = None
        self.status = Status.NEW
        self.scheduler = BackgroundScheduler()

    def setup(self, name: str, config: Config, scopes: list[Napalm]):
        """
        Set up the policy runner.

        Args:
        ----
            name: Policy name.
            config: Configuration data containing site information.
            scopes: scope data for the devices.

        """
        self.name = name.replace('\r\n', '').replace('\n', '')
        self.config = config

        if self.config is None:
            self.config = Config(defaults={})
        elif self.config.defaults is None:
            self.config.defaults = {}

        self.scheduler.start()
        for scope in scopes:
            sanitized_hostname = scope.hostname.replace('\r\n', '').replace('\n', '')
            if scope.driver and scope.driver not in supported_drivers:
                self.scheduler.shutdown()
                raise Exception(
                    f"Policy {self.name}, Hostname {sanitized_hostname}: specified driver '{scope.driver}' "
                    f"was not found in the current installed drivers list: {supported_drivers}."
                )

            if self.config.schedule is not None:
                logger.info(
                    f"Policy {self.name}, Hostname {sanitized_hostname}: Scheduled to run with '{self.config.schedule}'"
                )
                trigger = CronTrigger.from_crontab(self.config.schedule)
            else:
                logger.info(
                    f"Policy {self.name}, Hostname {sanitized_hostname}: One-time run"
                )
                trigger = DateTrigger(run_date=datetime.now() + timedelta(seconds=1))

            id = str(uuid.uuid4())
            self.scopes[id] = scope
            self.scheduler.add_job(
                self.run, id=id, trigger=trigger, args=[id, scope, self.config]
            )

            self.status = Status.RUNNING

    def run(self, id: str, scope: Napalm, config: Config):
        """
        Run the device driver code for a single scope item.

        Args:
        ----
            id: Job ID.
            scope: scope data for the device.
            config: Configuration data containing site information.

        """
        sanitized_hostname = scope.hostname.replace('\r\n', '').replace('\n', '')
        if scope.driver is None:
            logger.info(
                f"Policy {self.name}, Hostname {sanitized_hostname}: Driver not informed, discovering it"
            )
            scope.driver = discover_device_driver(scope)
            if scope.driver is None:
                self.status = Status.FAILED
                logger.error(
                    f"Policy {self.name}, Hostname {sanitized_hostname}: Not able to discover device driver"
                )
                try:
                    self.scheduler.remove_job(id)
                except Exception as e:
                    logger.error(
                        f"Policy {self.name}, Hostname {sanitized_hostname}: Error removing job: {e}"
                    )
                return

        logger.info(
            f"Policy {self.name}, Hostname {sanitized_hostname}: Get driver '{scope.driver}'"
        )

        try:
            np_driver = get_network_driver(scope.driver)
            logger.info(
                f"Policy {self.name}, Hostname {sanitized_hostname}: Getting information"
            )
            with np_driver(
                scope.hostname,
                scope.username,
                scope.password,
                scope.timeout,
                scope.optional_args,
            ) as device:
                data = {
                    "driver": scope.driver,
                    "device": device.get_facts(),
                    "interface": device.get_interfaces(),
                    "interface_ip": device.get_interfaces_ip(),
                    "defaults": config.defaults,
                }
                Client().ingest(scope.hostname, data)
        except Exception as e:
            logger.error(f"Policy {self.name}, Hostname {sanitized_hostname}: {e}")

    def stop(self):
        """Stop the policy runner."""
        self.scheduler.shutdown()
        self.status = Status.FINISHED



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""Device Discovery entry point."""

import argparse
import os
import sys
from importlib.metadata import version

import netboxlabs.diode.sdk.version as SdkVersion
import uvicorn

from device_discovery.client import Client
from device_discovery.server import app
from device_discovery.version import version_semver


def main():
    """
    Main entry point for the Agent CLI.

    Parses command-line arguments and starts the backend.
    """
    parser = argparse.ArgumentParser(description="Orb Device Discovery Backend")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"Device Discovery version: {version_semver()}, NAPALM version: {version('napalm')}, "
        f"Diode SDK version: {SdkVersion.version_semver()}",
        help="Display Device Discovery, NAPALM and Diode SDK versions",
    )
    parser.add_argument(
        "-s",
        "--host",
        default="0.0.0.0",
        help="Server host",
        type=str,
        required=False,
    )
    parser.add_argument(
        "-p",
        "--port",
        default=8072,
        help="Server port",
        type=int,
        required=False,
    )
    parser.add_argument(
        "-t",
        "--diode-target",
        help="Diode target",
        type=str,
        required=True,
    )

    parser.add_argument(
        "-k",
        "--diode-api-key",
        help="Diode API key. Environment variables can be used by wrapping them in ${} (e.g. ${MY_API_KEY})",
        type=str,
        required=True,
    )

    parser.add_argument(
        "-a",
        "--diode-app-name-prefix",
        help="Diode producer_app_name prefix",
        type=str,
        required=False,
    )

    try:
        args = parser.parse_args()
        api_key = args.diode_api_key
        if api_key.startswith("${") and api_key.endswith("}"):
            env_var = api_key[2:-1]
            api_key = os.getenv(env_var, api_key)

        client = Client()
        client.init_client(
            prefix=args.diode_app_name_prefix, target=args.diode_target, api_key=api_key
        )
        uvicorn.run(
            app,
            host=args.host,
            port=args.port,
        )
    except (KeyboardInterrupt, RuntimeError):
        pass
    except Exception as e:
        sys.exit(f"ERROR: Unable to start discovery backend: {e}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""NetBox Labs - Client Unit Tests."""

from types import SimpleNamespace
from unittest.mock import patch

import pytest

from device_discovery.client import Client
from device_discovery.translate import translate_data


@pytest.fixture
def sample_data():
    """Sample data for testing ingestion."""
    return {
        "device": {
            "hostname": "router1",
            "model": "ISR4451",
            "vendor": "Cisco",
            "serial_number": "123456789",
            "site": "New York",
            "driver": "ios",
        },
        "interface": {
            "GigabitEthernet0/0": {
                "is_enabled": True,
                "mtu": 1500,
                "mac_address": "00:1C:58:29:4A:71",
                "speed": 1000,
                "description": "Uplink Interface",
            }
        },
        "interface_ip": {
            "GigabitEthernet0/0": {"ipv4": {"192.0.2.1": {"prefix_length": 24}}}
        },
        "driver": "ios",
        "defaults": SimpleNamespace(site="New York", role=None, tags = None, device = None),
    }


@pytest.fixture
def mock_version_semver():
    """Mock the version_semver function."""
    with patch("device_discovery.client.version_semver", return_value="0.0.0") as mock:
        yield mock


@pytest.fixture
def mock_diode_client_class():
    """Mock the DiodeClient class."""
    with patch("device_discovery.client.DiodeClient") as mock:
        yield mock


def test_init_client(mock_diode_client_class, mock_version_semver):
    """Test the initialization of the Diode client."""
    client = Client()
    client.init_client(
        prefix="prefix", target="https://example.com", api_key="dummy_api_key"
    )

    mock_diode_client_class.assert_called_once_with(
        target="https://example.com",
        app_name="prefix/device-discovery",
        app_version=mock_version_semver(),
        api_key="dummy_api_key",
    )


def test_ingest_success(mock_diode_client_class, sample_data):
    """Test successful data ingestion."""
    client = Client()
    client.init_client(prefix="", target="https://example.com", api_key="dummy_api_key")

    mock_diode_instance = mock_diode_client_class.return_value
    mock_diode_instance.ingest.return_value.errors = []
    hostname = sample_data["device"]["hostname"]

    with patch(
        "device_discovery.client.translate_data",
        return_value=translate_data(sample_data),
    ) as mock_translate_data:
        client.ingest(hostname, sample_data)
        mock_translate_data.assert_called_once_with(sample_data)
        mock_diode_instance.ingest.assert_called_once()


def test_ingest_failure(mock_diode_client_class, sample_data):
    """Test data ingestion with errors."""
    client = Client()
    client.init_client(
        prefix="prefix", target="https://example.com", api_key="dummy_api_key"
    )

    mock_diode_instance = mock_diode_client_class.return_value
    mock_diode_instance.ingest.return_value.errors = ["Error1", "Error2"]
    hostname = sample_data["device"]["hostname"]

    with patch(
        "device_discovery.client.translate_data",
        return_value=translate_data(sample_data),
    ) as mock_translate_data:
        client.ingest(hostname, sample_data)
        mock_translate_data.assert_called_once_with(sample_data)
        mock_diode_instance.ingest.assert_called_once()

    assert len(mock_diode_instance.ingest.return_value.errors) > 0


def test_ingest_without_initialization():
    """Test ingestion without client initialization raises ValueError."""
    Client._instance = None  # Reset the Client singleton instance
    client = Client()
    with pytest.raises(ValueError, match="Diode client not initialized"):
        client.ingest("", {})



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""NetBox Labs - Discovery Unit Tests."""

import logging
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pytest
from napalm.base.base import NetworkDriver

from device_discovery.discovery import (
    discover_device_driver,
    napalm_driver_list,
    set_napalm_logs_level,
    supported_drivers,
    walk_napalm_packages,
)


@pytest.fixture
def mock_get_network_driver():
    """Mock the get_network_driver function from napalm."""
    with patch("device_discovery.discovery.get_network_driver") as mock:
        yield mock


@pytest.fixture
def mock_packages_distributions():
    """Mock the importlib.metadata.packages_distributions function."""
    with patch("device_discovery.discovery.packages_distributions") as mock:
        yield mock


@pytest.fixture
def mock_loggers():
    """Mock the logging.getLogger function for various loggers."""
    with patch("device_discovery.discovery.logging.getLogger") as mock:
        yield mock


@pytest.fixture
def mock_import_module():
    """Fixture to mock import_module."""
    with patch("device_discovery.discovery.import_module") as mock_import:
        yield mock_import


@pytest.fixture
def mock_walk_packages():
    """Fixture to mock walk_packages."""
    with patch("device_discovery.discovery.walk_packages") as mock_import:
        yield mock_import


def test_discover_device_driver_success(mock_get_network_driver):
    """
    Test successful discovery of a NAPALM driver.

    Args:
    ----
        mock_get_network_driver: Mocked get_network_driver function.

    """
    mock_driver_instance = MagicMock()
    mock_driver_instance.get_facts.return_value = {"serial_number": "ABC123"}

    mock_get_network_driver.side_effect = [
        MagicMock(return_value=mock_driver_instance)
    ] * len(supported_drivers)

    info = SimpleNamespace(
        hostname="testhost",
        username="testuser",
        password="testpass",
        timeout=10,
        optional_args={},
    )

    driver = discover_device_driver(info)
    assert driver in supported_drivers, "Expected one of the supported drivers"


def test_discover_device_driver_no_serial_number(mock_get_network_driver):
    """
    Test discovery when no serial number is found.

    Args:
    ----
        mock_get_network_driver: Mocked get_network_driver function.

    """

    def side_effect():
        mock_driver_instance = MagicMock()
        mock_driver_instance.get_facts.return_value = {"serial_number": "Unknown"}
        return mock_driver_instance

    mock_get_network_driver.side_effect = side_effect

    info = SimpleNamespace(
        hostname="testhost",
        username="testuser",
        password="testpass",
        timeout=10,
        optional_args={},
    )

    driver = discover_device_driver(info)
    assert driver is None, "Expected no driver to be found"


def test_discover_device_driver_exception(mock_get_network_driver):
    """
    Test discovery when exceptions are raised.

    Args:
    ----
        mock_get_network_driver: Mocked get_network_driver function.

    """
    mock_get_network_driver.side_effect = Exception("Connection failed")

    info = SimpleNamespace(
        hostname="testhost",
        username="testuser",
        password="testpass",
        timeout=10,
        optional_args={},
    )

    driver = discover_device_driver(info)
    assert driver is None, "Expected no driver to be found due to exception"


def test_discover_device_driver_mixed_results(mock_get_network_driver):
    """
    Test discovery with mixed results from drivers.

    Args:
    ----
        mock_get_network_driver: Mocked get_network_driver function.

    """

    def side_effect(driver_name):
        if driver_name == "nxos":
            mock_driver_instance = MagicMock()
            mock_driver_instance.get_facts.return_value = {"serial_number": "ABC123"}
            return mock_driver_instance
        raise Exception("Connection failed")

    mock_get_network_driver.side_effect = side_effect

    info = SimpleNamespace(
        hostname="testhost",
        username="testuser",
        password="testpass",
        timeout=10,
        optional_args={},
    )

    driver = discover_device_driver(info)
    assert driver == "nxos", "Expected the 'ios' driver to be found"


def test_napalm_driver_list(mock_packages_distributions, mock_import_module):
    """
    Test the napalm_driver_list function to ensure it correctly lists available NAPALM drivers.

    Args:
    ----
        mock_packages_distributions: Mocked importlib.metadata.packages_distributions function.
        mock_import_module: Mocked import_module function.

    """
    mock_distributions = [
        "napalm_srl",
        "napalm_fake_driver",
    ]

    class MockDriver(NetworkDriver):
        pass

    mock_module = MagicMock()
    setattr(mock_module, "MockDriver", MockDriver)
    mock_import_module.return_value = mock_module

    mock_packages_distributions.return_value = mock_distributions

    expected_drivers = ["ios", "eos", "junos", "nxos", "srl", "fake_driver"]
    drivers = napalm_driver_list()
    assert drivers == expected_drivers, f"Expected {expected_drivers}, got {drivers}"


def test_napalm_driver_list_error(mock_packages_distributions, mock_import_module):
    """
    Test the napalm_driver_list function when an error occurs during driver import.

    Args:
    ----
        mock_packages_distributions: Mocked importlib.metadata.packages_distributions function.
        mock_import_module: Mocked import_module function.

    """
    mock_distributions = [
        "napalm_srl",
    ]

    mock_import_module.side_effect = Exception("Import failed")
    mock_packages_distributions.return_value = mock_distributions
    expected_drivers = ["ios", "eos", "junos", "nxos"]

    with patch("device_discovery.discovery.logger") as mock_logger:
        drivers = napalm_driver_list()
        mock_logger.error.assert_called_once_with(
            f"Error importing module {mock_distributions[0]}: Import failed"
        )
        assert (
            drivers == expected_drivers
        ), f"Expected {expected_drivers}, got {drivers}"


def test_napalm_driver_list_nested(mock_packages_distributions, mock_import_module):
    """
    Test the napalm_driver_list function when a driver is found in a nested module.

    Args:
    ----
        mock_packages_distributions: Mocked importlib.metadata.packages_distributions function.
        mock_import_module: Mocked import_module function.

    """
    mock_distributions = [
        "napalm_srl",
    ]

    mock_module = MagicMock()
    mock_import_module.return_value = mock_module

    mock_packages_distributions.return_value = mock_distributions

    expected_drivers = ["ios", "eos", "junos", "nxos", "srl.nested"]

    with patch(
        "device_discovery.discovery.walk_napalm_packages", return_value=expected_drivers
    ):
        drivers = napalm_driver_list()
        assert (
            drivers == expected_drivers
        ), f"Expected {expected_drivers}, got {drivers}"


def test_walk_napalm_packages_success(mock_import_module, mock_walk_packages):
    """
    Test walk_napalm_packages function with valid modules.

    Args:
    ----
        mock_import_module: Mocked import_module function.
        mock_walk_packages: Mocked walk_packages function.

    """

    class MockDriver(NetworkDriver):
        pass

    mock_module = MagicMock()
    setattr(mock_module, "MockDriver", MockDriver)
    mock_module.__path__ = ["mock/path"]
    mock_module.__name__ = "napalm_test"
    mock_package = MagicMock()
    mock_package.name = "napalm_test.test_driver"

    mock_import_module.return_value = mock_module
    mock_walk_packages.return_value = [mock_package]

    result = walk_napalm_packages(mock_module, "napalm_", ["ios"])
    assert result == ["ios", "test.test_driver"]


def test_walk_napalm_packages_no_drivers(mock_import_module, mock_walk_packages):
    """
    Test walk_napalm_packages function when no valid drivers are found.

    Args:
    ----
        mock_import_module: Mocked import_module function.
        mock_walk_packages: Mocked walk_packages function.

    """
    mock_module = MagicMock()
    mock_module.__path__ = ["mock/path"]
    mock_module.__name__ = "napalm"
    mock_package = MagicMock()
    mock_package.name = "napalm.invalid_driver"

    mock_import_module.return_value = MagicMock()
    mock_walk_packages.return_value = [mock_package]

    with patch("device_discovery.discovery.inspect.getmembers", return_value=[]):
        result = walk_napalm_packages(mock_module, "napalm.", [])
        assert result == [], f"Expected an empty list, got {result}"


def test_walk_napalm_packages_exception_handling(
    mock_import_module, mock_walk_packages
):
    """
    Test walk_napalm_packages function with exceptions during module import.

    Args:
    ----
        mock_import_module: Mocked import_module function.
        mock_walk_packages: Mocked walk_packages function.

    """
    mock_module = MagicMock()
    mock_module.__path__ = ["mock/path"]
    mock_module.__name__ = "napalm"
    mock_package = MagicMock()
    mock_package.name = "napalm.error_driver"

    mock_import_module.side_effect = Exception("Import failed")
    mock_walk_packages.return_value = [mock_package]

    with patch("device_discovery.discovery.logger") as mock_logger:
        result = walk_napalm_packages(mock_module, "napalm.", [])
        mock_logger.error.assert_called_once_with(
            f"Error importing module {mock_package.name}: Import failed"
        )
        assert result == [], f"Expected an empty list, got {result}"


def test_set_napalm_logs_level(mock_loggers):
    """
    Test setting the logging level for NAPALM and related libraries.

    Args:
    ----
        mock_loggers: Mocked loggers for various libraries.

    """
    set_napalm_logs_level(logging.DEBUG)

    for logger in mock_loggers.values():
        logger.setLevel.assert_called_once_with(logging.DEBUG)
